#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#define NO_TST_MAIN
#include "tst.c"

struct file {
  char *dir,*name;
  time_t time;
  mode_t mode;
  nlink_t nlink;
  off_t size;
};

const char* months[] = {
  "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
};

void lstime(time_t T,char *dest) {
  struct tm *t;
  time_t now;
  time(&now);
  t=localtime(&T);
  if (T<=now && T>=now-60*60*12*365)
    sprintf(dest,"%s %2d %02d:%02d",months[t->tm_mon],t->tm_mday,t->tm_hour,t->tm_min);
  else
    sprintf(dest,"%s %2d  %04d",months[t->tm_mon],t->tm_mday,t->tm_year+1900);
}

void lsmode(mode_t t,char *dest) {
  sprintf(dest,"%c%c%c%c%c%c%c%c%c%c",
	  S_ISLNK(t)?'l':S_ISDIR(t)?'d':'-',
	  t&0400?'r':'-',
	  t&0200?'w':'-',
	  t&0100?'x':'-',
	  t&040?'r':'-',
	  t&020?'w':'-',
	  t&010?'x':'-',
	  t&04?'r':'-',
	  t&02?'w':'-',
	  t&01?'x':'-');
}

struct tst *init_dir(const char *path) {
  struct tst *t=0;
  struct stat s;
  DIR *dir=opendir(".");
  struct dirent *d;
  struct file *f;
  char *fullpath,*curpath;
  int uses=0;
  if (path)
    curpath=strdup(path);
  else
    curpath=strdup("/");
  for (;;) {
    d=readdir(dir);
    if (!d) break;
    if (lstat(d->d_name,&s))
      continue;
    if (!S_ISLNK(s.st_mode) &&
	!S_ISREG(s.st_mode) &&
	!S_ISDIR(s.st_mode))
      continue;
    if (d->d_name[0] == '.')
      continue;
    uses++;
    f=(struct file*)malloc(sizeof(struct file));
    f->dir=curpath;
    f->name=strdup(d->d_name);
    f->time=s.st_mtime;
    f->mode=s.st_mode;
    f->nlink=s.st_nlink;
    f->size=s.st_size;
    {
      char *tmp=alloca(strlen(curpath)+strlen(d->d_name)+3);
      sprintf(tmp,"%s%s",curpath,d->d_name);
      t=jointst(t,maketst(tmp,(void *)f));
/*      printf("Found %s\n",tmp); */
    }
    if (S_ISDIR(s.st_mode)) {
      if (path) {
	fullpath=(char *)malloc(strlen(path)+strlen(d->d_name)+3);
	sprintf(fullpath,"%s%s/",path,d->d_name);
      } else {
	fullpath=(char *)malloc(strlen(d->d_name)+3);
	sprintf(fullpath,"/%s/",d->d_name);
      }
      if (!chdir(d->d_name)) {
	t=jointst(t,init_dir(fullpath));
	chdir("..");
      }
      free(fullpath);
    }
#if 0
    lsmode(s.st_mode,mode);
    lstime(s.st_mtime,date);
    printf("%s %3d %-8s %-8s %8d %s %s\n",
	   mode,s.st_nlink,"r00t","rewt",s.st_size,date,d->d_name);
#endif
  }
  closedir(dir);
  if (!uses) free(curpath);
  return t;
}

void list(void *v) {
  struct file *f=(struct file *)v;
  char date[13];
  char mode[10];
  lsmode(f->mode,mode);
  lstime(f->time,date);
  printf("%s %3d %-8s %-8s %8d %s %s%s\n",
	  mode,f->nlink,"r00t","rewt",f->size,date,f->dir,f->name);
}

void globapplytst(struct tst *t,void (*foo)(void *),const char *s) {
  if (!t) return;
#if 0
  printf("globapplytst(%c,%s)\n",t->c,s);
  if (t->payload) {
    printf("mismatch: ");
    foo(t->payload);
  }
#endif
  if (!*s) return;
  switch (*s) {
  case '?':
    if (t->left) globapplytst(t->left,foo,s);
    if (t->c=='/') {
      if (t->right) globapplytst(t->right,foo,s);
      break;
    }
    if (s[1]==0) {
      if (t->payload)
	foo(t->payload);
    } else
      if (t->match)
	globapplytst(t->match,foo,s+1);
    if (t->right) globapplytst(t->right,foo,s);
    break;
  case '*':
    if (t->left) globapplytst(t->left,foo,s);
    if (t->c=='/')
      if (s[1]!='*') {
	if (t->right) globapplytst(t->right,foo,s);
	break;
      }
    if (s[1]==0 || (s[1]=='*' && s[2]==0)) {
      if (t->payload)
	foo(t->payload);
      if (t->match)
	globapplytst(t->match,foo,s);
    } else {
/*      printf("Nicht am Pattern-Ende.\n"); */
      if (t->match)
	globapplytst(t->match,foo,s);
    }
    if (s[1]!=0 || (s[1]=='*' && s[2]!=0)) {
      if (s[1]=='*') {
	if (t->match)
	  globapplytst(t->match,foo,s+2);
      } else {
	if (t->match)
	  globapplytst(t->match,foo,s+1);
      }
    }
    if (t->right) globapplytst(t->right,foo,s);
    break;
  default:
    if (*s==t->c) {
      if (s[1]==0)
	if (t->payload)
	  foo(t->payload);
      if (t->match)
	globapplytst(t->match,foo,s+1);
    } else if (*s<t->c) {
      if (t->left) globapplytst(t->left,foo,s);
    } else /* if (*s>t->c) */ {
      if (t->right) globapplytst(t->right,foo,s);
    }
  }
}


static struct file *array=0;
static int alloced=0;
static int used=0;

void append(void *v) {
  struct file *f=(struct file *)v;
  ++used;
  if (used>alloced) {
    alloced+=1024;
    array=realloc(array,alloced*sizeof(struct file));
  }
  array[used-1]=*f;
}

int sortSize(const void *a,const void *b) {
  struct file *A=(struct file *)a;
  struct file *B=(struct file *)b;
  return (B->size-A->size);
}

int sortTime(const void *a,const void *b) {
  struct file *A=(struct file *)a;
  struct file *B=(struct file *)b;
  return (B->time-A->time);
}

char *CanonifyFilename(const char *arg,const char *path) {
  char *canonify,*tmp;
/*  while (path[0]=='/') path++; */
  canonify=(char *)malloc(strlen(path)+strlen(arg)+5);
  if (arg[0]=='/')
    strcpy(canonify,arg);
  else
    sprintf(canonify,"%s/%s",path,arg);
  do {
    tmp=strstr(canonify,"//");
    if (tmp)
      strcpy(tmp,tmp+1);
  } while (tmp);
  do {
    if (canonify[0]=='.') {
      fprintf(stderr,"/.. requested\n");
      return 0;
    }
    tmp=strstr(canonify,"/.");
    if (tmp) {
      if (!strncmp(tmp,"/../",4)) {
	char *tmp2=tmp-1;
	while (tmp2>canonify && *tmp2!='/') tmp2--;
	if (tmp2==canonify) tmp++;
	strcpy(tmp2,tmp+3);
      } else if (!strcmp(tmp,"/..")) {
	char *tmp2=tmp-1;
	while (tmp2>canonify && *tmp2!='/') tmp2--;
/*	if (tmp2==canonify) tmp++; */
	strcpy(tmp2,tmp+3);
      } else {
	fprintf(stderr,"dot file requested\n");
	return 0;
      }
      printf("/%s\n",canonify);
    }
  } while (tmp);
  if (canonify[0]==0) {
    canonify[0]='/';
    canonify[1]=0;
  }
  return canonify;
}

const char *skipdir(const char *arg,const char *path) {
  if (!strncmp(arg,path,strlen(path))) {
    arg+=strlen(path);
    if (*arg=='/') arg++;
  }
  return arg;
}

char* dolist(struct tst *t,const char *arg,const char *path) {
  char *canonify;
  char *tmp;
  char *result=0;
  char Reverse=0,Long=0,bySize=0,byTime=0,Directory=0,noGroup=0;
  /* 0. Schritt: Optionen filtern */
  while (*arg==' ') arg++;
  while (*arg=='-') {
    arg++;
    while (*arg != ' ' && *arg && *arg != '\t') {
      switch (*arg++) {
      case 'l': Long = 1; break;
      case 'r': Reverse = 1; break;
      case 't': byTime = 1; break;
      case 'S': bySize = 1; break;
      case 'd': Directory = 1; break;
      case 'G': noGroup = 1; break;
      }
    }
    while (*arg == ' ' || *arg == '\t') arg++;
  }
  /* 1. Schritt: s,[^/]+/../,, */
/*  while (path[0]=='/') path++; */
  canonify=(char *)malloc(strlen(path)+strlen(arg)+5);
  if (arg[0]=='/')
    strcpy(canonify,arg);
  else
    sprintf(canonify,"%s/%s",path,arg);
  do {
    tmp=strstr(canonify,"//");
    if (tmp)
      strcpy(tmp,tmp+1);
  } while (tmp);
  {
    char *blub=canonify+strlen(canonify)-1;
    while (*blub=='/')
      *blub--=0;
    if (blub+1==canonify) {
      canonify[0]='/';
      canonify[1]='*';
      canonify[2]=0;
    }
  }
  do {
    if (canonify[0]=='.') {
      fprintf(stderr,"/.. requested\n");
      return 0;
    }
    tmp=strstr(canonify,"/.");
    if (tmp) {
      if (!strncmp(tmp,"/../",4)) {
	char *tmp2=tmp-1;
	while (tmp2>canonify && *tmp2!='/') tmp2--;
	if (tmp2==canonify) tmp++;
	strcpy(tmp2,tmp+3);
      } else if (!strcmp(tmp,"/..")) {
	char *tmp2=tmp-1;
	while (tmp2>canonify && *tmp2!='/') tmp2--;
	if (tmp2==canonify) tmp++;
	strcpy(tmp2,tmp+3);
      } else {
	fprintf(stderr,"dot file requested\n");
	return 0;
      }
      printf("/%s\n",canonify);
    }
  } while (tmp);
  printf("Canonified: %s\n",canonify);
  /* 2. Schritt: matchende Dateien extrahieren */
/*  fprintf(stderr,"Listing \"%s\"\n",canonify); */
  free(array); array=0; used=0; alloced=0;
  globapplytst(t,append,canonify);
  /* 3. Schritt: wenn nur eine Datei matcht, und diese Datei ein
   * Directory ist, dann "/*" anhngen und nochmal durchlaufen */
  if (used==1) {
    if (S_ISDIR(array[0].mode)) {
      char *blub=canonify+strlen(canonify);
      *blub++='/';
      *blub++='*';
      *blub++=0;
      free(array); array=0; used=0; alloced=0;
      globapplytst(t,append,canonify);
    }
  }
  /* 4. Schritt: In char* ausgeben */
  {
    int start=0,end=used,inc=1;
    int i,size=0,Used=1;
    char buf[2048];
    if (bySize)
      qsort(array,used,sizeof(struct file),sortSize);
    else if (byTime)
      qsort(array,used,sizeof(struct file),sortTime);
    if (Reverse) {
      start=used-1; end=-1; inc=-1;
    }
    for (i=start; i!=end; i+=inc) {
      struct file *f=array+i;
      if (Long) {
	char date[13];
	char mode[10];
	lsmode(f->mode,mode);
	lstime(f->time,date);
	if (noGroup)
	  sprintf(buf,"%s %3d %-8s %8d %s %s%s\r\n",
		  mode,f->nlink,"ftp",f->size,date,skipdir(f->dir,path),f->name);
	else
	  sprintf(buf,"%s %3d %-8s %-8s %8d %s %s%s\r\n",
		  mode,f->nlink,"ftp","ftp",f->size,date,skipdir(f->dir,path),f->name);
      } else {
	sprintf(buf,"%s%s\r\n",skipdir(f->dir,path),f->name);
      }
      Used+=strlen(buf);
      if (Used>size) {
	char *old=result;
	while (size<Used)
	  size+=4096;
	result=realloc(result,size);
	if (!old)
	  *result=0;
      }
      strcat(result,buf);
    }
  }
  return result;
}

#ifndef NO_DIR_MAIN
main() {
  struct tst *t;
  char *result;
  chdir("/ftp");
  t=init_dir(0);
  if (result=dolist(t,"pub","/")) {
    puts(result);
  } else {
    puts("list error");
  }
}
#endif
