#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

enum ftp_states {
  NEXT,
  START,
  ARG,

  FINISHED,

  RECOVER
};

enum ftp_commands {
  NONE,
  ABOR, AUTH, CDUP, CWD,
  DELE, LIST, MDTM, MKD,
  MODE, NLST, NOOP, PASS,
  PASV, PORT, PWD, QUIT,
  REST, RETR, RMD, RNFR,
  RNTO, SITE, SIZE, STOR,
  STRU, SYST, TYPE, USER,
  EPSV, EPRT
};

struct ftp_state {
  char buf[1024];
  int bufpos;
  char *inbuf;
  int start,end;
  char *outbuf;
  int ostart,olen;

  enum ftp_states state;
  enum ftp_commands command;
  char *arg1,*wd,cmd[10];
  int len1,cmdlen;	/* total length for both */
  int curlen;		/* currently read length for both */
  int type;
  int fd,datafd;
  int dataport;
  int passive;
  unsigned long rest;
#ifdef IPV6
  struct sockaddr_in6 peer,me;
#else
  struct sockaddr_in peer,me;
#endif
};

typedef enum ftp_states ftpstate;

ftpstate abor_callback(char *opaque_state);
ftpstate auth_callback(char *opaque_state);
ftpstate cdup_callback(char *opaque_state);
ftpstate cwd_callback(char *opaque_state);
ftpstate dele_callback(char *opaque_state);
ftpstate list_callback(char *opaque_state);
ftpstate mdtm_callback(char *opaque_state);
ftpstate mkd_callback(char *opaque_state);
ftpstate mode_callback(char *opaque_state);
ftpstate nlst_callback(char *opaque_state);
ftpstate noop_callback(char *opaque_state);
ftpstate pass_callback(char *opaque_state);
ftpstate pasv_callback(char *opaque_state);
ftpstate epsv_callback(char *opaque_state);
ftpstate eprt_callback(char *opaque_state);
ftpstate port_callback(char *opaque_state);
ftpstate pwd_callback(char *opaque_state);
ftpstate quit_callback(char *opaque_state);
ftpstate rest_callback(char *opaque_state);
ftpstate retr_callback(char *opaque_state);
ftpstate rmd_callback(char *opaque_state);
ftpstate rnfr_callback(char *opaque_state);
ftpstate rnto_callback(char *opaque_state);
ftpstate site_callback(char *opaque_state);
ftpstate size_callback(char *opaque_state);
ftpstate stor_callback(char *opaque_state);
ftpstate stru_callback(char *opaque_state);
ftpstate syst_callback(char *opaque_state);
ftpstate type_callback(char *opaque_state);
ftpstate user_callback(char *opaque_state);
ftpstate recover_callback(char *opaque_state);

struct argentry {
  char *s;
  ftpstate (*callback)(char *opaque_state);
} argtable[]=
  { {"abor",abor_callback},
    {"auth",auth_callback},
    {"cdup",cdup_callback},
    {"cwd", cwd_callback}, {"cwdt", cwd_callback}, {"xcwd", cwd_callback},
    {"dele",dele_callback},
    {"list",list_callback},
    {"mdtm",mdtm_callback},
    {"mkd", mkd_callback}, {"xmkd", mkd_callback},
    {"mode",mode_callback},
    {"nlst",nlst_callback},
    {"noop",noop_callback},
    {"pass",pass_callback},
    {"pasv",pasv_callback},
    {"port",port_callback},
    {"epsv",epsv_callback},
    {"eprt",eprt_callback},
    {"pwd", pwd_callback}, {"xpwd", pwd_callback},
    {"quit",quit_callback},
    {"rest",rest_callback},
    {"retr",retr_callback},
    {"rmd", rmd_callback}, {"xrmd", rmd_callback},
    {"rnfr",rnfr_callback},
    {"rnto",rnto_callback},
    {"site",site_callback},
    {"size",size_callback},
    {"stor",stor_callback},
    {"stru",stru_callback},
    {"syst",syst_callback},
    {"type",type_callback},
    {"user",user_callback}, {"user(",user_callback},
    {0,0},
  } ;


#define expect(what1,what2,new_state) \
  return (c==what1||c==what2?new_state:RECOVER)
#define expect1(what,new_state) \
  return (c==what?new_state:RECOVER)

enum ftp_states digest(char *opaque_state, char c, int fd) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
/*  fprintf(stderr,"Parsing: state %02x, input ",state->state);
  if (isprint(c))
    fprintf(stderr,"\"%c\"\n",c);
  else
    fprintf(stderr,"0x%d\n",c); */
  switch (state->state) {
  case NEXT:
    state->cmdlen=0;
    free(state->arg1); state->arg1=0;
    if (c=='\r' || c=='\n')
      return NEXT;
  case START:
    if (isalpha(c)) {
      if (state->cmdlen<10) {
	state->cmd[state->cmdlen++]=tolower(c);
	return START;
      }
    } else if (c==' ') {
      state->cmd[state->cmdlen]=0;
      if (state->arg1) free(state->arg1);
      state->len1=1024; state->arg1=(char *)malloc(1024); state->curlen=0;
      return ARG;
    } else if (c=='\r' || c=='\n') {
      struct argentry *tmp=argtable;
      state->cmd[state->cmdlen]=0;
#ifndef NDEBUG
      fprintf(stderr,"Command \"%s\"\n",state->cmd);
#endif
      for (tmp=argtable; tmp->s; tmp++)
	if (!strcmp(state->cmd,tmp->s))
	  return tmp->callback(opaque_state);
      return recover_callback(opaque_state);
    }
    return RECOVER;

  case ARG:
    if (c=='\r' || c=='\n') {
      struct argentry *tmp=argtable;
      state->arg1[state->curlen]=0;
#ifndef NDEBUG
      if (state->arg1)
	fprintf(stderr,"Command \"%s\" %s\n",state->cmd,state->arg1);
      else
	fprintf(stderr,"Command \"%s\"\n",state->cmd);
#endif
      for (tmp=argtable; tmp->s; tmp++)
	if (!strcmp(state->cmd,tmp->s))
	  return tmp->callback(opaque_state);
      return recover_callback(opaque_state);
    }
    if (state->curlen >= state->len1) {
      state->len1+=1024;
      state->arg1=realloc(state->arg1,state->len1);
      if (!state->arg1)
	return RECOVER;
    }
    state->arg1[state->curlen++]=c;
    return ARG;

  case FINISHED:
    return FINISHED;

  default:
    return RECOVER;
  }
}
