#define ATTACH_TO_STDIN

/* undef STATS if you don't want fftpd to collect statistics, might save
 * .01% CPU or something. */
#define STATS

/* todo:
 * Command "eprt" |1|10.0.0.4|32779|
 * Command epsv
 */

/* This would be much cleaner had I implemented it in C++.
 * Eventually someone ought to clean up this code. */

#ifdef __FreeBSD__
#define NO_ALLOCA_H
#endif

#define TOO_MANY_CACHED_ENTRIES 2

#define _XOPEN_SOURCE
#define __EXTENSIONS__
#define _BSD_SOURCE
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#ifndef NO_ALLOCA_H
#include <alloca.h>
#endif
#include <assert.h>

#define NO_DUMMY_CALLBACKS
#define NDEBUG
#include "proto.c"
#define NO_DIR_MAIN
#include "dir.c"

#include "ip4.h"
#include "ip6.h"

struct tst *filez=0;

#ifndef O_NDELAY
#ifdef O_NONBLOCK
#define O_NDELAY O_NONBLOCK
#else
#error neither O_NDELAY nor O_NONBLOCK is defined!  Please fix the includes.
#endif
#endif

fd_set rfd,wfd,_rfd,_wfd;
int max;	// highest file descriptor for select */

#ifdef STATS
unsigned long inbytes,outbytes,cinbytes,coutbytes,loutbytes;
unsigned long connections,fds;
#endif

const char *ftpprefix;

static int true=1;	/* needed for setsockopt :( */

/* Open a server socket, return a file descriptor ready for accept() or
 * select() or -1 on error.  The socket is set to non-blocking I/O and
 * address reuse. */
#ifdef IPV6
int server_socket_magic(short port,char addr[16]) {
#else
int server_socket_magic(short port,long addr) {
#endif
/* e.g.: port=80, addr=INADDR_ANY */
  int s;
#ifdef IPV6
  struct sockaddr_in6 sa;
  if ((s=socket(PF_INET6,SOCK_STREAM,IPPROTO_TCP))==-1) {
#else
  struct sockaddr_in sa;
  if ((s=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==-1) {
#endif
    perror("socket");
    return -1;
  }
  if (fcntl(s,F_SETFL,O_NDELAY)==-1) {
    perror("fcntl");
    return -1;
  }
  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(void *)&true, sizeof(true))==-1) {
    perror("setsockopt(SO_REUSEADDR)");
    return -1;
  }
#ifdef IPV6
  sa.sin6_family=PF_INET;
  sa.sin6_port=htons(port);
  memcpy(&sa.sin6_addr,addr,16);
#else
  sa.sin_family=AF_INET;
  sa.sin_port=htons(port);
  sa.sin_addr.s_addr=htonl(addr);
#endif
  if (bind(s,(struct sockaddr *)&sa,sizeof(sa))==-1) {
    perror("bind");
    return -1;
  }
  if (listen(s,100)==-1) {
    perror("listen");
    return -1;
  }
  return s;
}

/* Now comes the generic select() based TCP server part.
 * Each TCP connection is represented by a struct tcp_connection, which
 * knows the file descriptor, a protocol specific state (represented as
 * (char*)), flags and callbacks for reading, writing and timeout.
 * While a select() based server has the potential to be very efficient,
 * it also forces us to use non-blocking I/O and a state machine to
 * parse the protocol.  That makes the protocol handlers difficult to
 * write, hard to understand and ugly to debug. */

/* These are bits that can be set in the flags: */

/* Server sockets work like this: you do a server_socket_magic and
 * select on that file descriptor for reading.  When select indicates
 * data, you do an accept() on that socket and get a file descriptor.
 * You can then use the same socket again to accept the next connection.
 * The file descriptor returned by accept is then used to select for
 * reading and/or writing.  So we need a way to tell the server code
 * that this is not a file descriptor but a socket waiting for a
 * connection, and that flag is called CLONEME. */
#define CLONEME 1
/* This bit is set before the first socket. The server code uses it to
 * detect if it should call getpeername() to get the remote IP address
 * or just read data. */
#define INITIAL 2
/* The callbacks have to indicate whether the protocol now wants to read
 * or write data or both.  This is done by returning WANTREAD|WANTWRITE. */
#define WANTREAD 4
#define WANTWRITE 8

struct tcp_connection {
  int socket_fd;
  time_t timeout,kickat;
  char *state;  /* application specific */
  int flags;	/* CLONEME -> connection is automatically accepted again */
  int (*read_callback)(int fd);
  int (*write_callback)(int fd);
  int (*timeout_callback)(int fd);
};

/* Keep track of the file descriptors */
struct tcp_connection root[FD_SETSIZE];

/* basic initialization for struct tcp_connection */
void init_connection(struct tcp_connection *c) {
  c->socket_fd=-1;
  c->timeout=300;	/* 5 minutes */
  c->timeout=1800;	/* 30 minutes */
  c->kickat=0;
  c->state=NULL;
  c->flags=0;
  c->read_callback=c->write_callback=c->timeout_callback=0;
}

#ifdef INCLUDE_ECHO
/* I tested the generic server code with an echo server.
 * As you can see, the callback based mechanism makes the code ugly. */
struct echo_state {
  char buf[1024];
  int start,end;
};

int echo_read_callback(int fd) {
  struct echo_state *state;
  if (!root[fd].state)
    if (!(root[fd].state=(char *)malloc(sizeof(struct echo_state))))
      perror("malloc");
  state=(struct echo_state*)(root[fd].state);
  state->start=0;
  state->end=read(fd,state->buf,1024);
  printf("echo read callback on %d.  Read %d bytes.\n",fd,state->end);
  if (state->end<0) {
    perror("read");
    free(state); root[fd].state=0;
    close(fd);
    return 0;
  }
  if (state->end==0) {
    close(fd);
    free(state); root[fd].state=0;
    return 0;
  }
  return WANTWRITE;
}

int echo_write_callback(int fd) {
  struct echo_state *state=(struct echo_state*)(root[fd].state);
  int len=write(fd,state->buf+state->start,state->end-state->start);
  printf("echo write callback on %d.  Wrote %d bytes.\n",fd,len);
  if (len==0)
    return WANTREAD;
  state->start+=len;
  if (state->start>=state->end)
    return WANTREAD;
  else
    return WANTWRITE;
}

int echo_timeout_callback(int fd) {
  printf("echo timeout callback on %d.\n",fd);
  close(fd);
  free(root[fd].state); root[fd].state=0;
  return 0;
}
#endif


#ifdef ATTACH_TO_STDIN
/* For debugging, we also listen on stdin and kill the whole server if
 * we detect EOF. */
int control_read_callback(int fd) {
  char tmp[1024];
  int len=read(fd,tmp,1024);
  if (len==0)
    exit(0);
  return WANTREAD;
}
#endif




/* Output buffering routines.
 * We can't just write to the socket, because it is set to non-blocking
 * I/O.  That means we have to queue the data we want to write to the
 * socket.  This routine will move the data inside the buffer so that
 * there is no hole at the beginning. */
void compact_output_buffer(struct ftp_state *rs) {
  if (rs->ostart && rs->olen)
    memcpy(rs->outbuf,rs->outbuf+rs->ostart,rs->olen);
  rs->ostart=0;
  rs->outbuf=(char *)realloc(rs->outbuf,rs->olen+1);
}

/* Queue a write request.  Append the data to the output buffer,
 * reallocating it to make the data fit. */
void queue_write(struct ftp_state *rs, char *message, int len) {
  compact_output_buffer(rs);
  if (!len)
    return;
  rs->outbuf=(char *)realloc(rs->outbuf,rs->olen+len);
  memcpy(rs->outbuf+rs->olen,message,len);
  rs->olen+=len;
}

/* Write as much of the output buffer as we can. */
void do_partial_write(struct ftp_state *rs, int fd) {
  if (rs->olen) {
    int len=write(fd,rs->outbuf+rs->ostart,rs->olen);
    if (len>=0) {
#ifdef STATS
      coutbytes+=len;
#endif
      rs->olen-=len;
      rs->ostart+=len;
    }
  }
}

/* Convenience wrapper, write a string */
void queue_cstring(struct ftp_state *rs, char *message) {
  queue_write(rs,message,strlen(message));
}







/* FTP LIST (ls) callbacks */
struct ftplist_state {
  char *buf;
  int start,end;
  int ftpcmdfd;
};

int ftplist_read_callback(int fd) {
  printf("ftplist_write_callback called!1!!\n");
  return WANTWRITE;
}

int ftplist_write_callback(int fd) {
  struct ftplist_state *state=(struct ftplist_state*)(root[fd].state);
  int len;
  if (state->start==state->end) {	// Nothing to write
    struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
    free(state->buf);
    queue_cstring(f,"200 that's all.\r\n");
    root[state->ftpcmdfd].flags |= WANTWRITE;
    FD_SET(state->ftpcmdfd,&_wfd);
    if (close(fd)) perror("close1");
#if 0
    FD_CLR(state->fd,&_rfd);
    FD_CLR(state->fd,&_wfd);
#endif
    root[fd].flags=0;
#if 0
    close(state->fd);
#endif
    free(root[fd].state);
    root[fd].state=0;
    f->datafd=0;
    return 0;
  }
  if (state->end-state->start)
    len=write(fd,state->buf+state->start,state->end-state->start);
  printf("ftplist write callback on %d.  Wrote %d bytes.\n",fd,len);
  if (len<=0) {		// write error
    struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
    queue_cstring(f,"450 write error during list.\r\n");
    root[state->ftpcmdfd].flags |= WANTWRITE;
    if (close(fd)) perror("close2");
#if 0
    FD_CLR(state->fd,&_rfd);
    FD_CLR(state->fd,&_wfd);
#endif
    root[fd].flags=0;
#if 0
    close(state->fd);
#endif
    free(root[fd].state);
    root[fd].state=0;
    f->datafd=0;
    return 0;
  }
#ifdef STATS
  loutbytes+=len;
#endif
  state->start+=len;
  return WANTWRITE;
}

int ftplist_timeout_callback(int fd) {
  struct ftplist_state *state=(struct ftplist_state*)(root[fd].state);
  struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
  free(state->buf);
  printf("ftplist timeout callback on %d.\n",fd);
  queue_cstring(f,"450 timeout during file listing.\r\n");
  root[state->ftpcmdfd].flags |= WANTWRITE;
  if (close(fd)) perror("close3");
#if 0
  FD_CLR(state->fd,&_rfd);
  FD_CLR(state->fd,&_wfd);
#endif
  root[fd].flags=0;
#if 0
  close(state->fd);
#endif
  free(root[fd].state);
  root[fd].state=0;
  return 0;
}


/* FTP send (download) callbacks */
struct ftpsend_state {
  char buf[1024];
  int start,end;
  int ftpcmdfd;
  int fd;
};

int ftpsend_read_callback(int fd) {
  printf("ftpsend_read_callback called!1!!\n");
  return WANTWRITE;
}

int ftpsend_write_callback(int fd) {
  struct ftpsend_state *state=(struct ftpsend_state*)(root[fd].state);
  int len=0;
  if (state->start==state->end) {	// Nothing to write
    state->start=0;
    state->end=len=read(state->fd,state->buf,1024);
/*    printf("read %d bytes from downloaded file...\n",len); */
    if (len==0) {
      struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
      root[state->ftpcmdfd].kickat=time(0)+root[state->ftpcmdfd].timeout;
      queue_cstring(f,"200 download OK\r\n");
      root[state->ftpcmdfd].flags |= WANTWRITE;
      FD_SET(state->ftpcmdfd,&_wfd);
      if (close(fd)) perror("close4");
#if 0
      FD_CLR(state->fd,&_rfd);
      FD_CLR(state->fd,&_wfd);
#endif
      root[fd].flags=0;
      if (close(state->fd)) perror("close5");
      state->fd=0;
      free(root[fd].state);
      root[fd].state=0;
      f->datafd=0;
      return FINISHED;
    }
  }
  if (state->end-state->start)
    len=write(fd,state->buf+state->start,state->end-state->start);
/*  printf("ftpsend write callback on %d.  Wrote %d bytes.\n",fd,len); */
  if (len<=0) {		// write error
    struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
    queue_cstring(f,"450 write error during download.\r\n");
    root[state->ftpcmdfd].flags |= WANTWRITE;
    if (close(fd)) perror("close6");
#if 0
    FD_CLR(state->fd,&_rfd);
    FD_CLR(state->fd,&_wfd);
#endif
    root[fd].flags=0;
    if (close(state->fd)) perror("close7");
    free(root[fd].state);
    root[fd].state=0;
    return 0;
  }
#ifdef STATS
  outbytes+=len;
#endif
  state->start+=len;
  return WANTWRITE;
}

int ftpsend_timeout_callback(int fd) {
  struct ftpsend_state *state=(struct ftpsend_state*)(root[fd].state);
  struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
  printf("ftpsend timeout callback on %d.\n",fd);
  queue_cstring(f,"450 timeout during download.\r\n");
  root[state->ftpcmdfd].flags |= WANTWRITE;
  if (close(fd)) perror("close8");
#if 0
  FD_CLR(state->fd,&_rfd);
  FD_CLR(state->fd,&_wfd);
#endif
  root[fd].flags=0;
  if (close(state->fd)) perror("close9");
  free(root[fd].state);
  root[fd].state=0;
  return 0;
}





/* FTP retrieve (upload) callbacks */
struct ftprecv_state {
  int ftpcmdfd;	/* TODO */
  int fd;
  struct file f;
};

int ftprecv_read_callback(int fd) {
  char buf[1024];
  int len;
  struct ftprecv_state *state=(struct ftprecv_state*)(root[fd].state);
/*  printf("Reading from file at fd %d...\n",state->fd); */
  len=read(fd,buf,1024);
  if (len<=0) {	// EOF or ERROR
    struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
    root[state->ftpcmdfd].kickat=time(0)+root[state->ftpcmdfd].timeout;
    root[state->ftpcmdfd].flags |= WANTWRITE;
    FD_SET(state->ftpcmdfd,&_wfd);
    if (len<0)
      queue_cstring(f,"450 Planetary Interference is blocking transport beam, Sir!\r\n");
    else if (close(state->fd))
      queue_cstring(f,"450 Could not write file.  Disk full?\r\n");
    else
      queue_cstring(f,"200 Disengaging transporter beam.\r\n");
    /* TODO: insert the uploaded file in the in-code directory tree */
#if 0
    {
      struct stat s;
      fsync(fd);
      if (fstat(fd,&s)) {
      }
      struct tst *tmp=
    }
#endif
#if 0
    FD_CLR(state->fd,&_rfd);
    FD_CLR(state->fd,&_wfd);
#endif
    root[fd].flags=0;
    free(root[fd].state);
    root[fd].state=0;
    return 0;
  } else {
    struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
    int written=write(state->fd,buf,len);
    if (written<len) {
      queue_cstring(f,"450 write error during upload\r\n");
      root[state->ftpcmdfd].flags |= WANTWRITE;
      FD_SET(state->ftpcmdfd,&_wfd);
      if (close(fd)) perror("close10");
      if (close(state->fd)) perror("close11");
      free(root[fd].state);
      root[fd].state=0;
      return 0;
    }
#ifdef STATS
    inbytes+=len;
#endif
  }
  return WANTREAD;
}

int ftprecv_write_callback(int fd) {
  printf("ftprecv_write_callback called!1!!\n");
  return WANTREAD;
}

int ftprecv_timeout_callback(int fd) {
  struct ftprecv_state *state=(struct ftprecv_state*)(root[fd].state);
  struct ftp_state *f=(struct ftp_state*)(root[state->ftpcmdfd].state);
  printf("ftprecv timeout callback on %d.\n",fd);
  root[state->ftpcmdfd].flags |= WANTWRITE;
  if (close(fd)) perror("close12");
  if (close(state->fd)) perror("close13");
  free(root[fd].state);
  root[fd].state=0;
  return 0;
}




int dataconnection(struct ftp_state *state) {
  if (state->datafd==0) {
    queue_cstring(state,"425 Please specify the destination coordinates!\r\n");
    return NEXT;
  }
  /* Wir knnen es ja mal mit einem Connect probieren... */
  if (state->passive) {
    struct sockaddr_in dataconn;
    int fodder=sizeof(dataconn);
    int fd=accept(state->datafd,(struct sockaddr*)&dataconn,&fodder);
/*    printf("Listening on port %d\n",state->dataport); */
    if (fd<0)
      if (errno!=EAGAIN) {
	queue_cstring(state,"425 accept failed.\r\n");
	if (close(state->datafd)) perror("close14");
	state->datafd=0;
	return NEXT;
      }
    queue_cstring(state,"150 listening...\r\n");
/*    printf("accept(%d) -> %d\n",state->datafd,fd); */
    if (close(state->datafd)) perror("close15");
    state->datafd=fd;
  } else {
#ifdef IPV6
    struct sockaddr_in6 dataconn;
#else
    struct sockaddr_in dataconn;
#endif
/*    printf("Connecting to %s port %d\n",inet_ntoa(state->peer.sin_addr.s_addr),state->dataport); */
    if (fcntl(state->datafd,F_SETFL,O_NDELAY)==-1) {
      queue_cstring(state,"425 fcntl failed.\r\n");
      if (close(state->datafd)) perror("close16");
      state->datafd=0;
      return NEXT;
    }
#ifdef IPV6
    memcpy(&dataconn.sin6_addr,&state->peer.sin6_addr,16);
    dataconn.sin6_port=htons(state->dataport);
    dataconn.sin6_family=PF_INET6;
#else
    dataconn.sin_addr.s_addr=state->peer.sin_addr.s_addr;
    dataconn.sin_port=htons(state->dataport);
    dataconn.sin_family=AF_INET;
#endif
    if (connect(state->datafd,(struct sockaddr*)&dataconn,sizeof(dataconn))) {
      if (errno!=EAGAIN && errno!=EINPROGRESS && errno!=0) {
	perror("connect");
	queue_cstring(state,"425 Could not open data connection.\r\n");
	if (close(state->datafd)) perror("close17");
	state->datafd=0;
	return NEXT;
      }
    }
/*    printf("connect returned\n"); */
    queue_cstring(state,"150 connecting...\r\n");
  }
  return FINISHED;
}



/* These are the callbacks the FTP protocol parser calls when it
 * identifies one of the supported commands. */
ftpstate abor_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate auth_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate cdup_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate cwd_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  char *newpath=CanonifyFilename(state->arg1,state->wd);
  if (newpath) {
    struct tst *t;
    fprintf(stderr,"cd %s\n",newpath);
    t=findtst(filez,newpath);
    if (!strcmp(newpath,"/")) {
      free(state->wd);
      state->wd=newpath;
      queue_cstring(state,"200 OK\r\n");
      return NEXT;
    }
    if (t) {
      if (t->payload && S_ISDIR(((struct file *)(t->payload))->mode)) {
	free(state->wd);
	state->wd=newpath;
	queue_cstring(state,"200 OK\r\n");
	return NEXT;
      } else {
	fprintf(stderr,"not a directory: %s\n",newpath);
      }
    }
    free(newpath);
  }
  queue_cstring(state,"425 not found: \"");
  queue_cstring(state,state->arg1);
  queue_cstring(state,"\"\r\n");
  return NEXT;
}
ftpstate dele_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate list_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  fprintf(stderr,"list %s\n",state->arg1?state->arg1:"");
  if (state->arg1) {
    char *tmp=(char *)malloc(strlen(state->arg1)+4);
    sprintf(tmp,"-l %s",state->arg1);
    free(state->arg1); state->arg1=tmp;
  } else {
    char *tmp=strdup("-l");
    free(state->arg1); state->arg1=tmp;
  }
  nlst_callback(opaque_state);
#if 0
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
#endif
}
ftpstate mdtm_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  struct tm *t1;
  int datafd=-1;
  if (state->arg1) {
    char *tmp=CanonifyFilename(state->arg1,state->wd);
    struct tst *t;
    if (!tmp) {
      queue_cstring(state,"401 file not found\r\n");
      return NEXT;
    }
    fprintf(stderr,"mdtm %s\n",tmp);
    t=findtst(filez,tmp);
    if (t) {
      if (t->payload) {
	struct file *f=(struct file *)(t->payload);
	t1 = gmtime(&(f->time));
	if (t1) {
	  char buf[50];
	  sprintf(buf,"213 %04d%02d%02d%02d%02d%02d\r\n",
		  t1->tm_year+1900,t1->tm_mon+1,t1->tm_mday,
		  t1->tm_hour,t1->tm_min,t1->tm_sec);
	  queue_cstring(state,buf);
	  return NEXT;
	}
      }
    }
    free(tmp);
  }
  queue_cstring(state,"402 file not found\r\n");
  return NEXT;
}
ftpstate mkd_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate mode_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate nlst_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  struct ftplist_state *f;
  char *result;
  if (state->arg1)
    result=dolist(filez,state->arg1,state->wd);
  else
    result=dolist(filez,"",state->wd);
  if (!result) {
    queue_cstring(state,"500 problem.\r\n");
    return NEXT;
  }
  if (dataconnection(state) == NEXT)
    return NEXT;
  /* MAGIC! */
  root[state->datafd].flags=WANTWRITE;
  root[state->datafd].read_callback=ftplist_read_callback;
  root[state->datafd].write_callback=ftplist_write_callback;
  root[state->datafd].timeout_callback=ftplist_timeout_callback;
  FD_SET(state->datafd,&_wfd);
  if (root[state->datafd].state)
    free(root[state->datafd].state);
  root[state->datafd].socket_fd=state->datafd;
  f=(struct ftplist_state *)malloc(sizeof(struct ftplist_state));
  memset(f,0,sizeof(*f));
  root[state->datafd].state=(char *)f;
  f->ftpcmdfd=state->fd;
  f->buf=result;
  f->start=0; f->end=strlen(f->buf);
  if (state->datafd>max) max=state->datafd;
  return NEXT;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate noop_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"200 If you say so...\r\n");
  return NEXT;
}
ftpstate pass_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"230 yeah, man, whatever you say, man.\r\n");
  return NEXT;
}

ftpstate pasv_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  int datafd,fodder;
#ifdef IPV6
  struct sockaddr_in6 dataconn;
  unsigned char ip[16];
#else
  struct sockaddr_in dataconn;
  unsigned int ip;
#endif
  unsigned short port;
  if (state->datafd) {
    if (close(state->datafd)) perror("close18");
    FD_CLR(state->datafd,&_rfd);
    FD_CLR(state->datafd,&_wfd);
    root[state->datafd].flags=0;
    free(root[state->datafd].state);
    root[state->datafd].state=0;
    state->datafd=0;
  }
#ifdef IPV6
  datafd=server_socket_magic(0,(char*)&state->me.sin6_addr);
#else
  datafd=server_socket_magic(0,ntohl(state->me.sin_addr.s_addr));
#endif
  if (datafd<0) {
    queue_cstring(state,"425 All transporter bays are in use, Sir!\r\n");
    return NEXT;
  }
  fodder=sizeof(dataconn);
  if (getsockname(datafd,(struct sockaddr*)&dataconn,&fodder)) {
    queue_cstring(state,"425 Can't get local coordinates, Sir!\r\n");
    if (close(datafd)) perror("close19");
    return NEXT;
  }
#ifdef IPV6
  memcpy(ip,&dataconn.sin6_addr,16);
  port=ntohs((unsigned short int)dataconn.sin6_port);
#else
  ip=ntohl(dataconn.sin_addr.s_addr);
  port=ntohs((unsigned short int)dataconn.sin_port);
#endif
  {
    char buf[1024];
#ifdef IPV6
    if (!memcmp(ip,V4mappedprefix,12))
      sprintf(buf,"227 Passive Mode OK (%d,%d,%d,%d,%d,%d)\r\n",
	      ip[12],ip[13],ip[14],ip[15],(port>>8)&255,port&255 );
    else
      sprintf(buf,"227 Passive Mode OK (127,555,555,555,%d,%d)\r\n",(port>>8)&255,port&255);
#else
    sprintf(buf,"227 Passive Mode OK (%d,%d,%d,%d,%d,%d)\r\n",
	    (ip>>24)&255,(ip>>16)&255,(ip>>8)&255,ip&255,
	    (port>>8)&255,port&255 );
#endif
    queue_cstring(state,buf);
/*    puts(buf); */
  }
  state->passive=1;
  state->dataport=port;
  state->datafd=datafd;
/*  queue_cstring(state,"500 not implemented.\r\n"); */
  return NEXT;
}

ftpstate port_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  int a1,a2,a3,a4,p1,p2;
#ifdef IPV6
  struct sockaddr_in6 dataconn;
#else
  struct sockaddr_in dataconn;
#endif
  int on,datafd;
  if (state->datafd) {
    if (close(state->datafd)) perror("close20");
    FD_CLR(state->datafd,&_rfd);
    FD_CLR(state->datafd,&_wfd);
    root[state->datafd].flags=0;
    free(root[state->datafd].state);
    root[state->datafd].state=0;
    state->datafd=0;
  }
  if (state->arg1) {
    if (6==sscanf(state->arg1,"%d,%d,%d,%d,%d,%d",
		  &a1, &a2, &a3, &a4, &p1, &p2 ) ) {
      unsigned int port=(p1 << 8)+p2;
#ifdef IPV6
      memcpy((char*)&dataconn.sin6_addr,V4mappedprefix,12);
      dataconn=state->me;
      dataconn.sin6_addr.s6_addr[12]=a1;
      dataconn.sin6_addr.s6_addr[13]=a2;
      dataconn.sin6_addr.s6_addr[14]=a3;
      dataconn.sin6_addr.s6_addr[15]=a4;
#else
      unsigned int ip=(a1 << 24) + (a2 << 16) + (a3 << 8) + a4;
#endif

#ifdef IPV6
      datafd=socket(PF_INET6,SOCK_STREAM,IPPROTO_TCP);
#else
      datafd=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
#endif
      if (datafd<0) {
	queue_cstring(state,"425 Can't make data socket\r\n");
	return NEXT;
      }
      on=1;
      if (setsockopt(datafd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on))<0) {
	queue_cstring(state,"425 setsockopt(SO_REUSEADDR) failed\r\n");
	return NEXT;
      }
#ifdef IPV6
      if (memcmp(&dataconn.sin6_addr,&state->peer.sin6_addr,16)) {
#else
      if (htonl(ip) != state->peer.sin_addr.s_addr) {
#endif
	queue_cstring(state,"425 Look ma, an FTP bounce attack!\r\n");
	if (close(datafd)) perror("close21");
	return NEXT;
      }
      dataconn=state->me;
#ifdef IPV6
      dataconn.sin6_port=htons(20);	/* FTP data connection port */
#else
      dataconn.sin_port=htons(20);	/* FTP data connection port */
#endif
      errno=0;
      if (bind(datafd,(struct sockaddr*)&dataconn,sizeof(dataconn))) {
#ifdef IPV6
	dataconn.sin6_port=0;
#else
	dataconn.sin_port=0;
#endif
	if (bind(datafd,(struct sockaddr*)&dataconn,sizeof(dataconn))) {
	  char buf[1024];
	  sprintf(buf,"425 bind failed (%s)\r\n",strerror(errno));
/*	  puts(buf); */
	  queue_cstring(state,buf);
	  return NEXT;
	}
      }
      state->datafd=datafd;
      state->dataport=port;
      state->passive=0;
      queue_cstring(state,"200 Yeah yeah, go ahead.\r\n");
      return NEXT;
    }
  }
  queue_cstring(state,"501 What an interesting interpretation of RFC959...\r\n");
  return NEXT;
}

ftpstate epsv_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  int datafd,fodder;
#ifdef IPV6
  struct sockaddr_in6 dataconn;
  unsigned char ip[16];
#else
  struct sockaddr_in dataconn;
  unsigned int ip;
#endif
  unsigned short port;
  if (state->datafd) {
    if (close(state->datafd)) perror("close22");
    FD_CLR(state->datafd,&_rfd);
    FD_CLR(state->datafd,&_wfd);
    root[state->datafd].flags=0;
    free(root[state->datafd].state);
    root[state->datafd].state=0;
    state->datafd=0;
  }
#ifdef IPV6
  datafd=server_socket_magic(0,(char*)&state->me.sin6_addr);
#else
  datafd=server_socket_magic(0,ntohl(state->me.sin_addr.s_addr));
#endif
  if (datafd<0) {
    queue_cstring(state,"425 All transporter bays are in use, Sir!\r\n");
    return NEXT;
  }
  fodder=sizeof(dataconn);
  if (getsockname(datafd,(struct sockaddr*)&dataconn,&fodder)) {
    queue_cstring(state,"425 Can't get local coordinates, Sir!\r\n");
    if (close(datafd)) perror("close23");
    return NEXT;
  }
#ifdef IPV6
  memcpy(ip,&dataconn.sin6_addr,16);
  port=ntohs((unsigned short int)dataconn.sin6_port);
#else
  ip=ntohl(dataconn.sin_addr.s_addr);
  port=ntohs((unsigned short int)dataconn.sin_port);
#endif
  {
    char buf[1024];
    sprintf(buf,"229 Passive Mode OK (|||%d|)\r\n",port);
    queue_cstring(state,buf);
/*    puts(buf); */
  }
  state->passive=1;
  state->dataport=port;
  state->datafd=datafd;
/*  queue_cstring(state,"500 not implemented.\r\n"); */
  return NEXT;
}



ftpstate eprt_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  char sep,proto;
  int a1,a2,a3,a4,p1,p2;
/* "eprt |1|10.0.0.4|32779|" */
#ifdef IPV6
  struct sockaddr_in6 dataconn;
#else
  struct sockaddr_in dataconn;
  long ip;
#endif
  int on,datafd;
  if (state->datafd) {
    if (close(state->datafd)) perror("close24");
    FD_CLR(state->datafd,&_rfd);
    FD_CLR(state->datafd,&_wfd);
    root[state->datafd].flags=0;
    free(root[state->datafd].state);
    root[state->datafd].state=0;
    state->datafd=0;
  }
  if (state->arg1) {
    unsigned short port;
    sep=*state->arg1;
    proto=state->arg1[1];
    if (state->arg1[2]!=sep) goto parseerror;
    if (proto=='1') {
      char *tmp;
      unsigned long Port;
      tmp=strchr(state->arg1+3,sep);
      if (!tmp) goto parseerror;
/*      *tmp=0; */
#ifdef IPV6
      memcpy((char*)&dataconn.sin6_addr,V4mappedprefix,12);
      if (state->arg1[3+ip4_scan(state->arg1+3,dataconn.sin6_addr.s6_addr+12)] != sep) goto parseerror;
#else
      {
	char buf[4];
	if (state->arg1[3+ip4_scan(state->arg1+3,buf)] != sep) goto parseerror;
	ip=*(long*)buf;
      }
#endif
      if (tmp[1+scan_ulong(tmp+1,&Port)] != sep) goto parseerror;
      if (Port>65535) goto parseerror;
      port=Port;
#ifdef IPV6
    } else if (proto=='2') {
      char *tmp=strchr(state->arg1+3,sep);
      unsigned long Port;
      if (!tmp) goto parseerror;
      if (state->arg1[3+ip6_scan(state->arg1+3,dataconn.sin6_addr.s6_addr)] != sep) goto parseerror;
      if (tmp[1+scan_ulong(tmp+1,&Port)] != sep) goto parseerror;
      if (Port>65535) goto parseerror;
      port=Port;
#endif
    } else goto parseerror;

#ifdef IPV6
    datafd=socket(PF_INET6,SOCK_STREAM,IPPROTO_TCP);
#else
    datafd=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
#endif
    if (datafd<0) {
      queue_cstring(state,"425 Can't make data socket\r\n");
      return NEXT;
    }
    on=1;
    if (setsockopt(datafd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on))<0) {
      queue_cstring(state,"425 setsockopt(SO_REUSEADDR) failed\r\n");
      return NEXT;
    }
#ifdef IPV6
    if (memcmp(&dataconn.sin6_addr,&state->peer.sin6_addr,16)) {
#else
    if (htonl(ip) != state->peer.sin_addr.s_addr) {
#endif
      queue_cstring(state,"425 Look ma, an FTP bounce attack!\r\n");
      if (close(datafd)) perror("close25");
      return NEXT;
    }
    dataconn=state->me;
#ifdef IPV6
    dataconn.sin6_port=htons(20);	/* FTP data connection port */
#else
    dataconn.sin_port=htons(20);	/* FTP data connection port */
#endif
    errno=0;
    if (bind(datafd,(struct sockaddr*)&dataconn,sizeof(dataconn))) {
#ifdef IPV6
      dataconn.sin6_port=0;
#else
      dataconn.sin_port=0;
#endif
      if (bind(datafd,(struct sockaddr*)&dataconn,sizeof(dataconn))) {
	char buf[1024];
	sprintf(buf,"425 bind failed (%s)\r\n",strerror(errno));
/*	  puts(buf); */
	queue_cstring(state,buf);
	return NEXT;
      }
    }
    state->datafd=datafd;
    state->dataport=port;
    state->passive=0;
    queue_cstring(state,"200 Yeah yeah, go ahead.\r\n");
    return NEXT;
  }
parseerror:
  queue_cstring(state,"501 What an interesting interpretation of RFC2428...\r\n");
  return NEXT;
}


ftpstate pwd_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"257 \"");
  queue_cstring(state,state->wd);
  queue_cstring(state,"\"\r\n");
  return NEXT;
}
ftpstate quit_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"221 Winners never quit.  Quitters never win.\r\n");
  return FINISHED;
}
ftpstate rest_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  if (state->arg1) {
    if (1==sscanf(state->arg1,"%lu",&state->rest)) {
      queue_cstring(state,"350 ok.\r\n");
      return NEXT;
    }
  }
  queue_cstring(state,"501 invalid number.\r\n");
  return NEXT;
}
ftpstate retr_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  struct ftpsend_state *f;
  int datafd=-1;
  {
    char *tmp=CanonifyFilename(state->arg1,state->wd);
    struct tst *t;
    if (!tmp) {
      queue_cstring(state,"401 file not found\r\n");
      return NEXT;
    }
/*    fprintf(stderr,"get %s\n",tmp); */
    t=findtst(filez,tmp);
    if (t) {
      if (t->payload && S_ISREG(((struct file *)(t->payload))->mode)) {
	char *tmp2=(char *)alloca(strlen(ftpprefix)+strlen(tmp)+2);
	sprintf(tmp2,"%s%s",ftpprefix,tmp);
	datafd=open(tmp2,O_RDONLY);
	if (state->rest)
	  lseek(datafd,state->rest,SEEK_SET);
	state->rest=0;
      } else {
	fprintf(stderr,"not a regular file: %s\n",tmp);
      }
    }
    free(tmp);
  }
  if (datafd<0) {
    queue_cstring(state,"402 file not found\r\n");
    return NEXT;
  }
  if (dataconnection(state) == NEXT) {
    if (close(datafd)) perror("close26");
    return NEXT;
  }
  /* MAGIC! */
  root[state->datafd].flags=WANTWRITE;
  root[state->datafd].read_callback=ftpsend_read_callback;
  root[state->datafd].write_callback=ftpsend_write_callback;
  root[state->datafd].timeout_callback=ftpsend_timeout_callback;
  FD_SET(state->datafd,&_wfd);
  if (root[state->datafd].state)
    free(root[state->datafd].state);
  root[state->datafd].socket_fd=state->datafd;
  f=(struct ftpsend_state *)malloc(sizeof(struct ftpsend_state));
  memset(f,0,sizeof(*f));
  root[state->datafd].state=(char *)f;
  f->ftpcmdfd=state->fd;
  f->fd=datafd;
  f->start=0; f->end=0;
  if (state->datafd>max) max=state->datafd;
  return NEXT;
}
ftpstate rmd_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate rnfr_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate rnto_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate site_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate size_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  struct tm *t1;
  int datafd=-1;
  if (state->arg1) {
    char *tmp=CanonifyFilename(state->arg1,state->wd);
    struct tst *t;
    if (!tmp) {
      queue_cstring(state,"401 file not found\r\n");
      return NEXT;
    }
    fprintf(stderr,"size %s\n",tmp);
    t=findtst(filez,tmp);
    if (t) {
      if (t->payload) {
	struct file *f=(struct file *)(t->payload);
	t1 = gmtime(&(f->time));
	if (t1) {
	  char buf[50];
	  sprintf(buf,"213 %ld\r\n",f->size);
	  queue_cstring(state,buf);
	  return NEXT;
	}
      }
    }
    free(tmp);
  }
  queue_cstring(state,"402 file not found\r\n");
  return NEXT;
}
ftpstate stor_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  struct ftprecv_state *f;
  int datafd=-1;
  {
    char *tmp=CanonifyFilename(state->arg1,state->wd);
    char *tmp2=(char *)alloca(strlen(ftpprefix)+strlen(tmp)+2);
    char *f;
    sprintf(tmp2,"%s%s",ftpprefix,tmp);
    f=strrchr(tmp2,'/');
    if (f) {
      struct stat s;
      *f=0;
      if (!stat(tmp2,&s)) {
	*f='/';
	if (s.st_mode&2) {
	  datafd=open(tmp2,O_WRONLY|O_TRUNC|O_CREAT,0622);
	  free(tmp);
	} else {
	  queue_cstring(state,"402 permission denied\r\n");
	  return NEXT;
	}
      }
    }
  }
  if (datafd<0) {
    queue_cstring(state,"402 file not found\r\n");
    return NEXT;
  }
  if (dataconnection(state) == NEXT) {
    if (close(datafd)) perror("close27");
    return NEXT;
  }
  /* MAGIC! */
  root[state->datafd].flags=WANTREAD;
  root[state->datafd].read_callback=ftprecv_read_callback;
  root[state->datafd].write_callback=ftprecv_write_callback;
  root[state->datafd].timeout_callback=ftprecv_timeout_callback;
  FD_SET(state->datafd,&_rfd);
  if (root[state->datafd].state)
    free(root[state->datafd].state);
  root[state->datafd].socket_fd=state->datafd;
  f=(struct ftprecv_state *)malloc(sizeof(struct ftprecv_state));
  memset(f,0,sizeof(*f));
  root[state->datafd].state=(char *)f;
  f->ftpcmdfd=state->fd;
  f->fd=datafd;
  /* TODO: pass destination directory */
#if 0
  f->f.dir=
#endif
  if (state->datafd>max) max=state->datafd;
  return NEXT;
}
ftpstate stru_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"500 not implemented.\r\n");
  return NEXT;
}
ftpstate syst_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"215 UNIX Type: L8\r\n");
  return NEXT;
}
ftpstate type_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  int type=0;
  type=0;
  if (state->arg1) {
    if (tolower(*state->arg1) == 'a')
      type=1;
    else if (tolower(*state->arg1) == 'i')
      type=2;
    else if (tolower(*state->arg1) == 'l')
      if (state->arg1[1] == '8')
	type=2;
      else if (isdigit(state->arg1[1]))
	queue_cstring(state,"504 I only support 8-bit bytes.\r\r");
      else
	type=2;
    if (type) {
      if (type==2)
	queue_cstring(state,"215 Yeah, yeah, binary mode.\r\n");
      else
	queue_cstring(state,"215 ASCII mode.  Were they on drugs when they invented this?!\r\n");
      state->type=type;
    }
    return NEXT;
  }
  queue_cstring(state,"501 What an interesting interpretation of RFC959...\r\n");
  return NEXT;
}
ftpstate user_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  state->rest=0;
  if (state->arg1) {
/*    puts(state->arg1); */
    if (!strcmp(state->arg1,"ftp") ||
	!strcmp(state->arg1,"anonymous")) {
      queue_cstring(state,"230 Welcome.  Please behave yourself.\r\n");
      return NEXT;
    }
  }
  queue_cstring(state,"500 This server implements only anonymous FTP.\r\n");
  return NEXT;
}

ftpstate recover_callback(char *opaque_state) {
  struct ftp_state *state=(struct ftp_state *)opaque_state;
  queue_cstring(state,"501 What client are you using?  Microsoft FTP?!\r\n");
  return NEXT;
}



void free_ftp_state(struct ftp_state *state) {
  if (state->datafd) if (close(state->datafd)) perror("close28");
  if (state->inbuf) free(state->inbuf);
  if (state->outbuf) free(state->outbuf);
  if (state->arg1) free(state->arg1);
  if (state->wd) free(state->wd);
}

int ftp_read_callback(int fd) {
  struct ftp_state *state;
  int len, i;
  if (!root[fd].state) {
    fprintf(stderr,"read callback on uninitialized state\n");
    exit(1);
  }
  state=(struct ftp_state*)(root[fd].state);
  state->start=0;
  len=read(fd,state->buf,1024);
#ifdef DEBUG
  printf("rdbm read callback on %d.  Read %d bytes.\n",fd,len);
#endif
  if (len<0) {
    perror("read");
    free_ftp_state(state);
    free(state); root[fd].state=0;
#ifdef STATS
    connections--;
#endif
    if (close(fd)) perror("close29");
    return 0;
  }
  if (len==0) {
#ifdef STATS
    connections--;
#endif
    if (close(fd)) perror("close30");
    free_ftp_state(state);
    free(state); root[fd].state=0;
    return 0;
  }
#ifdef STATS
  cinbytes+=len;
#endif
  state->inbuf=(char *)realloc(state->inbuf,state->end+len);
  memcpy(state->inbuf+state->end,state->buf,len);
  state->end+=len;
  for (i=state->start; i<state->end; i++) {
/*    putchar(state->inbuf[i]); */
    state->state=digest(root[fd].state,state->inbuf[i],fd);
    if (state->state==FINISHED) {
      printf("Closing connection on %d\n",fd);
#ifdef STATS
      connections--;
#endif
      if (close(fd)) perror("close31");
      free_ftp_state(state);
      free(state); root[fd].state=0;
      return 0;
    }
/*    printf("%d\n",state->state); */
  }
  state->end=state->start;
  if (state->olen)
    return (state->state==FINISHED?0:WANTREAD)+WANTWRITE;
  return (state->state==FINISHED?0:WANTREAD);
}

int ftp_write_callback(int fd) {
  struct ftp_state *state;
  if (!root[fd].state) {
#ifdef IPV6
    struct sockaddr_in6 si;
#else
    struct sockaddr_in si;
#endif
    int len=sizeof(si);
    if (getpeername(fd,(struct sockaddr *)&si,&len)) {
      perror("getpeername");
      write(fd,"500 getpeername failed\n",23);
#ifdef STATS
      connections--;
#endif
      if (close(fd)) perror("close32");
      return 0;
    }
/*    printf("connect from %s\n",inet_ntoa(si.sin_addr.s_addr)); */
    if (!(root[fd].state=(char *)malloc(sizeof(struct ftp_state)))) {
      write(fd,"500 Out of memory\r\n",19);
#ifdef STATS
      connections--;
#endif
      if (close(fd)) perror("close33");
      return 0;
    }
    state=(struct ftp_state*)(root[fd].state);
    state->olen=0; state->ostart=0; state->outbuf=state->inbuf=0;
    state->end=0;
    state->state=START; state->peer=si;
    len=sizeof(si);
    if (getsockname(fd,(struct sockaddr *)&si,&len)) {
      perror("getsockname");
      write(fd,"500 getsockname failed\r\n",24);
#ifdef STATS
      connections--;
#endif
      if (close(fd)) perror("close34");
      return 0;
    }
    state->me=si;
    state->arg1=0; state->curlen=state->cmdlen=0;
    state->type=2;
    state->wd=strdup("/");
    state->datafd=0; state->fd=fd;
    queue_cstring(state,"220 Fefe's FTP server OK\r\n");
  }
  state=(struct ftp_state*)(root[fd].state);
  do_partial_write(state,fd);
  if (state->olen)
    return (state->state==FINISHED?0:WANTREAD)+WANTWRITE;
  if (state->state==FINISHED) {
    /*    printf("Closing connection on %d\n",fd); */
#ifdef STATS
    connections--;
#endif
    if (close(fd)) perror("close35");
    free_ftp_state(state);
    free(state); root[fd].state=0;
    return 0;
  }
  return WANTREAD;
}

int ftp_timeout_callback(int fd) {
  struct ftp_state *state=(struct ftp_state*)(root[fd].state);
  printf("timeout callback on %d.\n",fd);
#ifdef STATS
  connections--;
#endif
  if (close(fd)) perror("close36");
  free(state->inbuf);
  free(state->outbuf);
  free(state); root[fd].state=0;
  return 0;
}


int main(int argc,char *argv[]) {
  int s;
  struct timeval tv;
#ifdef STATS
  time_t lastt;
#endif
  char *dir=0;
  struct passwd *pwd=getpwnam("ftp");
#ifdef IPV6
  char any[16];

  memset(any,0,16);
#endif

  signal(SIGPIPE,SIG_IGN);
  if (argc>1) {
    dir=argv[1];
  } else {
    if (!pwd) {
      fprintf(stderr,"ftp user not found!\n");
      exit(1);
    }
    dir=pwd->pw_dir;
  }
  if (chdir(dir)) {
    fprintf(stderr,"could not chdir to FTP home %s\n",dir);
    exit(1);
  }
  if (chroot(".")) {
    fprintf(stderr,"could not chroot to FTP home %s\n",dir);
    ftpprefix=dir;
  } else
    ftpprefix="";
  puts("dir init");
  filez=init_dir(0);
  puts("done");
  umask(022);

  for (s=0; s<FD_SETSIZE; s++)
    init_connection(&(root[s]));
  FD_ZERO(&_rfd);
  memcpy(&_wfd,&_rfd,sizeof(_wfd));
#ifdef IPV6
  s=server_socket_magic(21,any);
  if (s<0)
    s=server_socket_magic(2121,any);
#else
  s=server_socket_magic(21,INADDR_ANY);
  if (s<0)
    s=server_socket_magic(2121,INADDR_ANY);
#endif
  if (s<0) {
    printf("error %d, %s.\n",s,strerror(errno));
    exit(0);
  }
  root[s].flags=INITIAL+WANTWRITE+CLONEME;
  root[s].read_callback=ftp_read_callback;
  root[s].write_callback=ftp_write_callback;
  root[s].timeout_callback=ftp_timeout_callback;
  FD_SET(s,&_rfd);

#ifdef ATTACH_TO_STDIN
  FD_SET(0,&_rfd);
  root[0].read_callback=control_read_callback;
  root[0].timeout=0;
#endif
  max=s;
#ifdef STATS
  inbytes=outbytes=cinbytes=coutbytes=loutbytes=connections=0;
  fds=0;
  lastt=time(0);
#endif
  do {
    time_t t=time(0);
    memcpy(&wfd,&_wfd,sizeof(_wfd));
    memcpy(&rfd,&_rfd,sizeof(_rfd));
    tv.tv_sec=5;
    tv.tv_usec=0;
#if 0
    {
      int i;
      printf("Selecting for (max %d):\n",max);
      for (i=0; i<=max; i++) {
	int j=FD_ISSET(i,&rfd);
	int k=FD_ISSET(i,&wfd);
	if (j || k) {
	  printf("  %d (%c%c)\n",i,j?'r':'-',k?'w':'-');
	}
      }
    }
#endif
    if (select(max+1,&rfd,&wfd,0,&tv)) {
      int i;
      t=time(0);
#ifdef STATS
      if (t!=lastt) {
	lastt=t;
	fprintf(stderr,"out: %8lu data, %8lu control, %8lu list.  in: %8lu data, %8lu control.   %d connections\n",
		outbytes,coutbytes,loutbytes,inbytes,cinbytes,connections);
	inbytes=outbytes=cinbytes=coutbytes=loutbytes=0;
      }
#endif
      for (i=0; i<=max; i++) {
	if (FD_ISSET(i,&rfd)) {
	  if (root[i].flags & INITIAL) {
	    int fd;
	    struct sockaddr_in bla;
	    int socklen;
	    /*	    printf("Connection auf %d!\n",i); */
	    if ((root[i].flags & CLONEME) == 0)
	      FD_CLR(i,&_rfd);
	    socklen=sizeof(bla);
	    fd=accept(i,(struct sockaddr *)&bla,&socklen);
	    if (fd>=0) {
	      connections++;
	      /*	      printf("Connection akzeptiert -> %d.\n",fd); */
	      root[fd]=root[s];
	      root[fd].flags-=INITIAL;
	      root[fd].kickat=t+root[fd].timeout;
	      if (root[fd].flags & WANTREAD) FD_SET(fd,&_rfd);
	      if (root[fd].flags & WANTWRITE) FD_SET(fd,&_wfd);
	      if (fd>max) max=fd;
	    } else {
	      perror("accept");
	    }
	  } else {
	    int j=root[i].read_callback(i);
	    if (j&WANTREAD) FD_SET(i,&_rfd); else FD_CLR(i,&_rfd);
	    if (j&WANTWRITE) FD_SET(i,&_wfd); else FD_CLR(i,&_wfd);
	  }
	  if (root[i].timeout)
	    root[i].kickat=t+root[i].timeout;
	}
	if (FD_ISSET(i,&wfd)) {
	  int j;
	  j=root[i].write_callback(i);
	  if (root[i].timeout)
	    root[i].kickat=t+root[i].timeout;
	  if (j&WANTREAD) FD_SET(i,&_rfd); else FD_CLR(i,&_rfd);
	  if (j&WANTWRITE) FD_SET(i,&_wfd); else FD_CLR(i,&_wfd);
	}
      }
    }
    {
      int i;
      time_t t=time(0);
/*      printf("Reaping connections...\n"); */
#ifdef STATS
      fds=0;
#endif
      for (i=0; i<=max; i++)
	if (FD_ISSET(i,&_rfd) || FD_ISSET(i,&_wfd)) {
#ifdef STATS
	  fds++;
#endif
/*	  printf("What about %d [%d vs %d]?\n",i,(int)t,root[i].kickat); */
	  if (!(root[i].flags & INITIAL) && 
	      root[i].kickat && root[i].kickat<=t) {
	    int j=root[i].timeout_callback(i);
	    if (j&WANTREAD) FD_SET(i,&_rfd); else FD_CLR(i,&_rfd);
	    if (j&WANTWRITE) FD_SET(i,&_wfd); else FD_CLR(i,&_wfd);
/*	    printf("%d timed out\n",i);
	    close(i); */
	  }
	}
    }
  } while (1);
  return 0;
}
