/* Ternary Search Trees */

#include <string.h>
#include <alloca.h>

struct tst {
  struct tst *left,*right,*match;
  char c;
  void *payload;
};

void inittst(struct tst *t) {
  t->c=0;
  t->left=t->right=t->match=0;
  t->payload=0;
}

void killtst(struct tst *t) {
  if (t->left) killtst(t->left);
  if (t->right) killtst(t->right);
  if (t->match) killtst(t->match);
#ifndef NDEBUG
  inittst(t);
#endif
  free(t->payload);
  free(t);
}

struct tst *findtst(struct tst *t,const char *c) {
  if (!t) return 0;
  if (t->c==*c) {
    c++;
    if (*c)
      return findtst(t->match,c);
    else
      return t;
  }
  if (*c<t->c) return findtst(t->left,c);
  if (*c>t->c) return findtst(t->right,c);
}

struct tst *jointst(struct tst *t,struct tst *q) {
  if (!t) return q;
  if (!q) return t;
  if (q->left || q->right) {
    t=jointst(t,q->left);
    t=jointst(t,q->right);
    q->left=q->right=0;
  }
  if (q->c==t->c) {
    t->match=jointst(t->match,q->match);
#ifndef NDEBUG
    inittst(q);
#endif
    free(q);
    return t;
  } else if (q->c < t->c) {
    t->left=jointst(t->left,q);
    return t;
  } else {
    t->right=jointst(t->right,q);
    return t;
  }
}

struct tst *maketst(const char *s,void *payload) {
  struct tst *tmp;
  if (!s || !*s)
    return 0;
  tmp=(struct tst *)malloc(sizeof(struct tst));
  inittst(tmp);
  tmp->c=*s;
  tmp->match=maketst(s+1,payload);
  if (!tmp->match)
    tmp->payload=payload;
  return tmp;
}

void dumptst_r(struct tst *t,const char *c);

void dumptst_nr(struct tst *t,char *c,char *c2) {
  if (t->left) dumptst_nr(t->left,c,c2);
  *c2=t->c;
  if (t->payload)
    puts(c);
  dumptst_r(t->match,c);
  if (t->right) dumptst_nr(t->right,c,c2);
}

void dumptst_r(struct tst *t,const char *c) {
  char *tmp,*tmp2;
  if (!t) return;
  tmp=alloca(strlen(c)+2);
  strcpy(tmp,c);
  tmp2=tmp+strlen(c);
  tmp2[1]=0;
  dumptst_nr(t,tmp,tmp2);
}

void dumptst(struct tst *t) {
  dumptst_r(t,"");
}

void applytst(struct tst *t,void (*foo)(void *)) {
  if (!t) return;
  if (t->left) applytst(t->left,foo);
  if (t->payload)
    foo(t->payload);
  if (t->match) applytst(t->match,foo);
  if (t->right) applytst(t->right,foo);
}

#ifndef NO_TST_MAIN
main() {
  struct tst *t=maketst("foo",(char *)0xfefec0de);
  struct tst *t1=maketst("bar",(char *)0xfefec0de);
  struct tst *t2=jointst(t,t1);
  dumptst(t);
}
#endif
